/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *
 **/

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "socket_srv.h"
#include "my_stdio.h"
#include "iar_bus.h"


#define BAD_CMD -1
#define BAD_PARAMETER -2
#define BAD_CLIENT -3



#define PRINT_FMT_MESSAGE(fd, cm) write(fd, cm, sizeof(struct ib_cm_line))


struct ib_dev *init_ib_dev()
{
	struct ib_dev *cl;

	cl = (struct ib_dev *) malloc(sizeof(struct ib_dev));
	cl->fd = -1;
	cl->transfer = 1;
	cl->alive = 1;
	cl->id = NULL;
	//cl->cmd_ind = 0;
	cl->mesg = NULL;
	cl->next = cl->prev = NULL;
	
	return cl;
}


int ib_dev_addr(struct ib_dev *cl, int id)
{
	if (id == 0)	
		cl->addr = IB_BASE_ADDR | cl->fd ;
	else 
		cl->addr = id;

	return 0;
}


int free_ib_dev(struct ib_dev *cl)
{
	if(cl->id != NULL) free(cl->id);
	free(cl);

	return 0;
}


struct ib_dev *find_ib_clien (struct ib_dev *cl, char *to_id)
{
	struct ib_dev *clp = cl;
	
	while (clp->next != NULL && strcmp(clp->id, to_id) != 0){
		clp = clp->next;
	}
	if (clp->next == NULL) clp = NULL;
	return clp;
}



int print_bus_data(struct ib_cm_line *cm, struct ib_dev *cl, int all)
{
	struct ib_dev *clp = cl;	
	
	while(clp->next != NULL){
		if ( (clp->fd) && 
		     (clp->transfer || all) && 
		     (clp->alive) &&
		     ( clp->addr != cm->addr_f)  ) {
			if( !PRINT_FMT_MESSAGE(clp->fd, cm) ){
				perror("write client");
				clp->alive = 0;
			}
			    }
		clp = clp->next;
	}
	
	return 0;
}





static struct ib_command *find_ib_command(struct ib_command *cm, int cm_id)
{
	int i;

	for (i = 0; cm[i].cm_id; i++)
		if (cm[i].cm_id == cm_id )
			return (&cm[i]);
	
	return ((struct ib_command *)NULL);
}


/* Execute a command. */
static int execute_command(struct ib_command *cm, struct ib_dev *cl, 
			   struct ib_dev *current)
{
	struct ib_command *command;
	
	command = find_ib_command (cm, current->cmd.cm_id);

	if (!command) {
		return BAD_CMD;
	}

	/* Call the function. */
	return ((*(command->func)) (cl, current));
}



int command_from_dev(struct ib_command *cm, struct ib_dev *cl, 
		     struct ib_dev *clp)
{
	memset(&clp->cmd, 0, sizeof(struct ib_cm_line));
	if ( read(clp->fd, &clp->cmd, 
		  sizeof(struct ib_cm_line)) <= 0) { /* Ops... die */
		printf("read client fail\n");
		close(clp->fd);
		clp->alive = 0;
		return -1;
	}
	
	switch( execute_command(cm, cl, clp) ){
	case 0:
		break;
	default:
	case BAD_CMD:
		fd_printf(clp->fd,"cmd error\n");
		break;
	case BAD_PARAMETER:
		fd_printf(clp->fd,"bad parameter\n");
			break;
	case BAD_CLIENT:
		fd_printf(clp->fd,"bad client name\n");
		break;
	}

	return 0;
}


int init_iar_bus_main(struct ib_command *cm, int fd_main, int fd_socket )
{
	struct ib_dev *cl, *clp, *cld;
	fd_set rfds;
	struct timeval tv;
	int retval;
	int l;
	struct sockaddr isa;
	int addrlen;
	struct ib_cm_line cm_main;

	addrlen = sizeof(isa); /* find socket's address for accept() */
	getsockname(fd_socket, (struct sockaddr *)&isa, &addrlen);

	cm_main.addr_f = IB_MAIN_ADDR;
	cm_main.addr_t = IB_BROADCAST_ADDR;
	cm_main.cm_id = 0;

	cl = init_ib_dev();
	for(;1;){
		clp = cl ;
		FD_ZERO(&rfds);
		FD_SET(fd_socket, &rfds);
		if (fd_main >= 0) FD_SET(fd_main, &rfds);
		while(clp->next != NULL) {
			if (clp->fd && clp->alive ){ 
				FD_SET(clp->fd, &rfds);
				clp = clp->next;
			}
			else if ( !clp->alive ) { 
                                /* OK: client dead, cleaning */ 
				cld = clp;
				if (clp->prev != NULL){
					clp = cld->prev;
					clp->next = cld->next;
					cld->next->prev = clp;
					clp = clp->next;
				}else {
					clp = cl = clp->next;
					cl->prev = NULL;
				}
				free_ib_dev(cld);		
			}
		}
		tv.tv_sec = 10;
		tv.tv_usec = 0;
		/* FIXME: 50 ? implement this dynamic */
		retval = select(50, &rfds, NULL, NULL, &tv);
		if (retval) {
			clp = cl ;
			while(clp->next != NULL) {
				if( FD_ISSET(clp->fd, &rfds) && clp->alive ) {
					command_from_dev(cm, cl, clp); 
				}
				clp = clp->next;
			}
			if (FD_ISSET(fd_socket, &rfds)) {
				clp = cl ;
				while(clp->next != NULL) clp = clp->next;
				clp->fd = accept(fd_socket, 
						 (struct sockaddr *)&isa, 
						 &addrlen); 
				if( clp->fd ){
					clp->alive = 1;
					printf("connected\n");
					clp->next = init_ib_dev();
					clp->next->prev = clp;
					ib_dev_addr(clp, 0);
				}
			}
			if ((fd_main >= 0) && FD_ISSET(fd_main, &rfds)) {
				memset(&cm_main.arg,0, IB_COMMAND_LINE_LEGEND);
				l = read(fd_main, &cm_main.arg, 
					 IB_COMMAND_LINE_LEGEND);
				if (l > 0) {
					/* hay que formatear el
                                           mensage de main */
					print_bus_data(&cm_main, cl, 0);
				}
			}
		}
	}

	return 0;
}


int init_iar_bus_local(struct ib_command *cm, 
			    int fd_main, char *sock_name )
{
	int fd_socket;

	if ( (fd_socket = socket_init_local( sock_name )) < 0) {
		return -1;
	}

	return init_iar_bus_main(cm, fd_main, fd_socket );
}


int init_iar_bus_inet(struct ib_command *cm, int fd_main, int port )
{
	int fd_socket;

	if ( (fd_socket = socket_init( port )) < 0) {
		return -1;
	}

	return init_iar_bus_main(cm, fd_main, fd_socket);
}


/****************************************************************************/
